import base64
import io
from itertools import product

import pytest

import cydrogen
from cydrogen._basekey import BaseKey  # type: ignore[import]

KEY_BYTES = b"\xeaCU\x91\xdb?\xc2\xefZ\xb9HO\x84\xc5\xf3\xbf\x07\xecunR\xab\xa4\xa7`\r=\xe5\xf0e\xfb%"
KEY_BYTES_2 = b")\xf5G\x95\xa9e\xeb\xb3\xfcU1?[\x9f\xd1\x8b6\xe4\x8a\xac\xf1 \xf5\xd3\xf8\x98c\xae\xce\xdcoA"

MESSAGES = [
    b"Nobody inspects the spammish repetition",
    b"Then shalt thou count to three, no more, no less. Three shall be the number thou shalt count, and the number of the counting shall be three.",
    b"I don't want to talk to you no more, you empty headed animal food trough wiper.",
]


def test_create_zero_hash_key():
    key = cydrogen.HashKey()
    assert bytes(key) == 32 * b"\x00"
    assert key.is_zero()
    assert not bool(key)


def test_create_hash_key_from_bytes():
    key = cydrogen.HashKey(KEY_BYTES)
    assert bytes(key) == KEY_BYTES
    assert bool(key)


def test_create_hash_key_from_str():
    key_str = base64.standard_b64encode(KEY_BYTES).decode("utf-8")
    key = cydrogen.HashKey(key_str)
    assert str(key) == key_str
    assert bool(key)


def test_gen_hash_key():
    key = cydrogen.HashKey.gen()
    assert bool(key)
    assert len(bytes(key)) == 32


def test_copy_hash_key():
    key1 = cydrogen.HashKey.gen()
    key2 = cydrogen.HashKey(key1)
    assert bytes(key1) == bytes(key2)
    assert key1 == key2


def test_hash_key_equality():
    key1 = cydrogen.HashKey(KEY_BYTES)
    key2 = cydrogen.HashKey(KEY_BYTES)
    key3 = cydrogen.HashKey.gen()
    assert key1 == key1
    assert key1 == key2
    assert key1 != key3


def test_hash_key_repr():
    key_str = base64.standard_b64encode(KEY_BYTES).decode("utf-8")
    key = cydrogen.HashKey(KEY_BYTES)
    assert repr(key) == f"HashKey({repr(key_str)})"


def test_no_convert_hashkey_secretboxkey():
    key = cydrogen.HashKey(KEY_BYTES)
    with pytest.raises(TypeError):
        cydrogen.SecretBoxKey(key)


def test_no_convert_secretboxkey_hash_key():
    key = cydrogen.SecretBoxKey(KEY_BYTES)
    with pytest.raises(TypeError):
        cydrogen.HashKey(key)


def test_convert_basekey_hash_key():
    key = BaseKey.gen()
    hk = cydrogen.HashKey(key)
    assert bytes(hk) == bytes(key)


def test_no_convert_hash_key_basekey():
    key = cydrogen.HashKey(KEY_BYTES)
    with pytest.raises(TypeError):
        BaseKey(key)


def test_simple_digest():
    key = cydrogen.HashKey(KEY_BYTES)
    h = key.hasher(MESSAGES[0]).digest()
    expected_hash = b"\x95\x08\x00\x14T9\xdcnr\x94\xf5C$\xfb.?"
    assert len(h) == 16
    assert h == expected_hash


def test_digest():
    master_keys = [cydrogen.HashKey(), cydrogen.HashKey(KEY_BYTES), cydrogen.HashKey(KEY_BYTES_2)]
    ctxs = [cydrogen.Context(b"EXAMPLES"), cydrogen.Context(b"CONTEXTS")]
    sizes = [16, 32, 64, 128]
    digests = set()
    for msg, key, ctx, digest_size in product(MESSAGES, master_keys, ctxs, sizes):
        h = cydrogen.Hash(key=key, ctx=ctx, digest_size=digest_size)
        h.update(msg)
        digest = h.digest()
        assert len(digest) == digest_size
        h = key.hasher(msg, ctx=ctx, digest_size=digest_size)
        assert h.digest() == digest
        assert digest not in digests
        digests.add(digest)


def test_hash_file():
    key = cydrogen.HashKey(KEY_BYTES)
    buf = cydrogen.gen_random_buffer(70000)
    fobj = io.BytesIO(buf)

    # call function hash_file with default chunk size
    fobj.seek(0)
    digest1 = cydrogen.hash_file(fobj, ctx=b"CONTEXTS", digest_size=16, key=key, chunk_size=io.DEFAULT_BUFFER_SIZE)
    assert len(digest1) == 16
    # call function hash_file with custom chunk size
    fobj.seek(0)
    digest2 = cydrogen.hash_file(fobj, ctx=b"CONTEXTS", digest_size=16, key=key, chunk_size=65536)
    assert digest1 == digest2
    # call hasher method update_from
    fobj.seek(0)
    hasher = key.hasher(ctx=b"CONTEXTS")
    hasher.update_from(fobj)
    assert hasher.digest() == digest1



class CyException(Exception):
    """
    Base class for all exceptions in the `cydrogen` package.
    """

    ...

class EncryptException(CyException):
    """
    Exception raised for errors during encryption operations.
    """

    ...

class DecryptException(CyException):
    """
    Exception raised for errors during decryption operations.

    In particular, this exception is raised when the decryption fails due to an invalid key
    or tampered ciphertext.
    """

    ...

class DeriveException(CyException):
    """
    Exception raised for errors during key derivation operations.
    """

    ...

class SignException(CyException):
    """
    Exception raised for errors during signing operations.
    """

    ...

class VerifyException(CyException):
    """
    Exception raised for errors during signature verification operations.

    In particular, this exception is raised when the signature verification fails due to an
    invalid signature or tampered data.
    """

    ...



# cython: language_level=3

import base64
import io

from libc.stdint cimport uint64_t
from libc.stdint cimport uint32_t

from ._basekey cimport BaseKey
from ._context cimport make_context
from ._exceptions cimport DecryptException, EncryptException
from ._hash cimport Hash, HashKey
from ._masterkey cimport MasterKey, make_masterkey
from ._sign import SignPublicKey, SignSecretKey, SignKeyPair
from ._utils cimport FileOpener, SafeMemory, SafeReader, SafeWriter, TeeWriter
from ._utils cimport store64, load64, store32, load32

from ._decls cimport hydro_secretbox_HEADERBYTES, secretbox_encrypt, secretbox_decrypt


# declared as a cdef so that it cannot be modified at runtime
cdef bytes _ENC_MSG_HEADER = b"qN\x00\x00"

ENC_MSG_HEADER = _ENC_MSG_HEADER[:4]


cdef class SecretBoxKey(BaseKey):
    def __init__(self, key):
        if key is None:
            raise ValueError("Key argument cannot be None")

        if isinstance(key, SafeMemory):
            super().__init__(key)
            return

        # when key argument is already a SecretBoxKey, copy the key
        cdef SecretBoxKey o
        if isinstance(key, SecretBoxKey):
            o = <SecretBoxKey>key
            super().__init__(o.key)
            return

        if isinstance(key, (HashKey, MasterKey, SignKeyPair, SignPublicKey, SignSecretKey)):
            raise TypeError("can't create a SecretBoxKey from another concrete key type")

        # when key argument is a string, assume it's a base64 encoded key
        if isinstance(key, str):
            super().__init__(base64.standard_b64decode(key))
            return

        # else, assume it's a bytes like object
        super().__init__(bytes(key))

    def __eq__(self, other):
        if not isinstance(other, SecretBoxKey):
            return False
        cdef SecretBoxKey o = <SecretBoxKey>other
        return self.eq(o)

    def __repr__(self):
        return f'SecretBoxKey({repr(str(self))})'

    @classmethod
    def from_password(cls, const unsigned char[:] password, *, master_key=None, ctx=None, opslimit=10000):
        if password is None:
            raise ValueError("Password cannot be None")
        cdef mkey = make_masterkey(master_key)
        cdef BaseKey derived = mkey.derive_key_from_password(password, ctx=ctx, opslimit=opslimit)
        return cls(derived)

    cpdef secretbox(self, ctx=None):
        return SecretBox(self, ctx=ctx)


cdef make_secretbox_key(key):
    if isinstance(key, SecretBoxKey):
        return key
    return SecretBoxKey(key)


cdef class EncryptedMessage:
    def __init__(self, const unsigned char[:] ctext, uint64_t msg_id):
        if ctext is None:
            raise ValueError("Message cannot be None")
        self.ciphertext = bytes(ctext)
        self.msg_id = msg_id

    def __bytes__(self):
        w = io.BytesIO()
        self.writeto(w)
        return w.getvalue()

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, EncryptedMessage):
            return False
        cdef EncryptedMessage o = <EncryptedMessage>other
        if self.msg_id != o.msg_id:
            return False
        return self.ciphertext == o.ciphertext

    cpdef writeto(self, out):
        if out is None:
            raise ValueError("File object cannot be None")

        cdef bytearray header = bytearray(4 + 8 + 8)
        header[0:4] = _ENC_MSG_HEADER
        cdef unsigned char[:] header_view = header
        store64(header_view[4:12], len(self.ciphertext))
        store64(header_view[12:20], self.msg_id)

        cdef SafeWriter w = SafeWriter(out)
        cdef size_t n_written = w.write(header)
        n_written += w.write(self.ciphertext)
        cdef size_t expected_size = len(header) + len(self.ciphertext)
        if n_written < expected_size:
            raise IOError("Failed to write the entire message to the file object")
        return n_written

    @classmethod
    def from_bytes(cls, const unsigned char[:] framed):
        if framed is None:
            raise ValueError("Framed message cannot be None")
        return cls.read_from(io.BytesIO(framed))

    @classmethod
    def read_from(cls, reader, *, max_msg_size=None):
        if reader is None:
            raise ValueError("File object cannot be None")
        cdef SafeReader r = SafeReader(reader)

        cdef bytearray header_buf = bytearray(20)
        if r.readinto(header_buf) < 20U:
            raise IOError("Failed to read next message header")
        if header_buf[:4] != _ENC_MSG_HEADER:
            raise ValueError("Invalid message header")
        cdef size_t msg_size = load64(header_buf[4:12])
        if max_msg_size is not None:
            if msg_size > <size_t>max_msg_size:
                raise ValueError("Message size exceeds maximum allowed size, {} > {}".format(msg_size, max_msg_size))
        cdef uint64_t msg_id = load64(header_buf[12:20])
        cdef bytearray msg = bytearray(msg_size)
        if r.readinto(msg) != msg_size:
            raise IOError("Failed to read the entire message")
        return cls(msg, msg_id)

    cpdef decrypt(self, key, ctx=None, out=None):
        if key is None:
            raise ValueError("Key cannot be None")
        return SecretBox(key, ctx=ctx).decrypt(self.ciphertext, msg_id=self.msg_id, out=out)


cdef class SecretBox:
    def __init__(self, key, *, ctx=None):
        if key is None:
            raise ValueError("Key cannot be None")
        self.key = make_secretbox_key(key)
        self.ctx = make_context(ctx)

    cpdef encrypt(self, const unsigned char[:] plaintext, uint64_t msg_id=0, out=None):
        if plaintext is None:
            raise ValueError("Plaintext cannot be None")
        if out is not None:
            SafeWriter(out)  # ensure out is a file-like object
        cdef bytearray ciphertext = bytearray(len(plaintext) + hydro_secretbox_HEADERBYTES)
        try:
            secretbox_encrypt(plaintext, msg_id, self.ctx, self.key, ciphertext)
        except ValueError:
            raise
        except Exception as ex:
            raise EncryptException("Encryption failed") from ex
        cdef EncryptedMessage msg = EncryptedMessage(ciphertext, msg_id)
        if out is not None:
            # write the encrypted message to the output writer
            # note that EncryptedMessage.writeto writes the full frame including the header0
            msg.writeto(out)
        return msg.ciphertext

    cpdef decrypt(self, ciphertext, uint64_t msg_id=0, out=None):
        if ciphertext is None:
            raise ValueError("Ciphertext cannot be None")
        cdef SafeWriter w
        if out is not None:
            w = SafeWriter(out)

        if isinstance(ciphertext, EncryptedMessage):
            _id = ciphertext.msg_id
            ciphertext = ciphertext.ciphertext
            if msg_id != 0U and _id != msg_id:
                raise DecryptException("The passed message ID does not match the one in the EncryptedMessage")
            msg_id = _id

        if len(ciphertext) < hydro_secretbox_HEADERBYTES:
            raise ValueError("Ciphertext is too short")
        cdef size_t plaintext_len = len(ciphertext) - hydro_secretbox_HEADERBYTES
        cdef bytearray plaintext = bytearray(plaintext_len)
        try:
            secretbox_decrypt(ciphertext, msg_id, self.ctx, self.key, plaintext)
        except ValueError:
            raise
        except Exception as ex:
            raise DecryptException("Decryption failed") from ex
        if out is not None:
            w.write(plaintext)
        return bytes(plaintext)

    cpdef encrypt_file(self, src, dst, size_t chunk_size=io.DEFAULT_BUFFER_SIZE):
        if src is None or dst is None:
            raise ValueError("Source and destination file objects cannot be None")
        with FileOpener(src, mode="rb") as src_obj, FileOpener(dst, mode="wb") as dst_obj:
            return self._encrypt_file(src_obj, dst_obj, chunk_size=chunk_size)

    cdef _encrypt_file(self, fileobj, out, size_t chunk_size=io.DEFAULT_BUFFER_SIZE):
        if fileobj is None or out is None:
            raise ValueError("Source and destination file objects cannot be None")
        if chunk_size <= hydro_secretbox_HEADERBYTES:
            raise ValueError("Chunk size must be greater than the header size")
        SafeReader(fileobj)  # ensure fileobj is a reader
        cdef SafeWriter w = SafeWriter(out)
        cdef Hash hasher = Hash(ctx=self.ctx, key=bytes(self.key))
        cdef uint64_t total_bytes_written = 0
        cdef bytearray buf = bytearray(chunk_size - hydro_secretbox_HEADERBYTES)     # the buffer used to read a chunk of the file
        cdef unsigned char[:] buf_view = buf
        cdef size_t n = 0               # the size of the current plaintext chunk
        cdef uint64_t msg_id = 1        # we will increment this for each chunk

        # write the max buffer size to the output file so that we can read it at decrypt time
        cdef bytearray header = bytearray(8)
        header[0:4] = _ENC_MSG_HEADER
        cdef unsigned char[:] header_view = header
        store32(header_view[4:8], chunk_size)  # store the chunk size in the header
        w.write(header)  # write the header to the output file
        total_bytes_written += 8

        while True:
            n = fileobj.readinto(buf)               # read a chunk of the plaintext file into the plaintext buffer
            if n == 0:
                break
            hasher.update(buf_view[:n])
            self.encrypt(buf_view[:n], msg_id=msg_id, out=w)
            total_bytes_written += n + hydro_secretbox_HEADERBYTES + 20
            msg_id += 1

        # encrypt and write the hash of the original file
        self.encrypt(hasher.digest(), msg_id=0, out=w)
        return total_bytes_written

    cpdef decrypt_file(self, src, dst):
        if src is None or dst is None:
            raise ValueError("Source and destination file objects cannot be None")
        with FileOpener(src, mode="rb") as src_obj, FileOpener(dst, mode="wb") as out_obj:
            return self._decrypt_file(src_obj, out_obj)

    cdef _decrypt_file(self, fileobj, out):
        if fileobj is None or out is None:
            raise ValueError("Source and destination file objects cannot be None")
        cdef uint64_t msg_id = 1
        cdef uint64_t total_bytes_written = 0
        cdef bytearray sbuf = bytearray(8)
        cdef size_t max_buf_size = 0
        cdef EncryptedMessage enc_msg
        cdef Hash hasher = Hash(ctx=self.ctx, key=bytes(self.key))
        cdef TeeWriter tee
        cdef bytes transmitted_hash
        cdef bytes computed_hash

        cdef SafeReader r = SafeReader(fileobj)
        cdef SafeWriter w = SafeWriter(out)

        if (<uint32_t>r.readinto(sbuf)) != 8U:
            raise IOError("Failed to read max buffer size")
        if sbuf[:4] != _ENC_MSG_HEADER:
            raise ValueError("Invalid message header")
        max_buf_size = load32(sbuf[4:8])
        tee = TeeWriter(w, hasher)

        while True:
            try:
                enc_msg = EncryptedMessage.read_from(r, max_msg_size=max_buf_size)
            except IOError as ex:
                # we have reached the end of the file without having seen the hash
                raise DecryptException("final hash not found")
            if enc_msg.msg_id == 0:
                # if the message ID is 0, we assume it's the hash of the original file
                # we don't need to write it to the output file
                break
            if enc_msg.msg_id != msg_id:
                raise DecryptException("Invalid message ID")
            enc_msg.decrypt(self.key, ctx=self.ctx, out=tee)
            total_bytes_written += len(enc_msg.ciphertext) - hydro_secretbox_HEADERBYTES
            msg_id += 1

        # the last encrypted message contains hash of the original file
        # we need to compare it with the hash of the decrypted file
        transmitted_hash = enc_msg.decrypt(self.key, ctx=self.ctx)
        if len(transmitted_hash) != hasher.digest_size:
            raise DecryptException("Invalid hash length")
        computed_hash = hasher.digest()
        if transmitted_hash != computed_hash:
            raise DecryptException("Invalid hash")
        return total_bytes_written



# cython: language_level=3

cimport cython

from cpython.buffer cimport PyBuffer_FillInfo, PyBUF_WRITABLE
from libc.stdint cimport uint64_t
from libc.stdint cimport uint32_t
from libc.stdint cimport uint16_t

import io
import os
import pathlib
import tempfile


cdef class SafeMemory:
    def __cinit__(self, size_t size):
        if size == 0:
            return
        self.ptr = cyd_malloc(size)
        if self.ptr != NULL:
            self.size = size
            cyd_memzero(self.ptr, self.size)

    def __dealloc__(self):
        if self.ptr != NULL:
            cyd_free(self.ptr)

    def __init__(self, size_t size):
        if size == 0:
            raise ValueError("size must be greater than 0")
        if self.ptr == NULL:
            raise MemoryError("Failed to allocate memory")
        self.readonly_protected = 0

    def __len__(self):
        return self.size

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        if flags & PyBUF_WRITABLE:
            # consumer is asking for writable buffer
            if self.readonly_protected:
                raise ValueError("Memory is read-only, cannot provide writable buffer")
            PyBuffer_FillInfo(buffer, self, self.ptr, self.size, 0, flags)
        else:
            PyBuffer_FillInfo(buffer, self, self.ptr, self.size, 1, flags)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, SafeMemory):
            return False
        cdef SafeMemory o = <SafeMemory>other
        if self.size != o.size:
            return False
        if self.ptr == o.ptr:
            return True
        return cyd_memcmp(self.ptr, o.ptr, self.size) == 0

    def __bool__(self):
        return cyd_is_zero(<const unsigned char*>(self.ptr), self.size) == 0

    cdef set(self, const unsigned char[:] data):
        if self.readonly_protected == 1:
            raise ValueError("Memory is read-only")
        if data is None:
            raise ValueError("Data cannot be None")
        if len(data) != len(self):
            raise ValueError(f"Data must be {self.size} bytes long, got {len(data)} bytes")
        if len(data) == 0:
            self.mark_readonly()
            return
        cdef unsigned char[:] view = self
        view[0:len(self)] = data[0:len(self)]
        del view
        self.mark_readonly()

    cdef set_zero(self):
        if self.readonly_protected == 1:
            raise ValueError("Memory is read-only")
        cyd_memzero(self.ptr, self.size)
        self.mark_readonly()

    cdef mark_readonly(self):
        if self.readonly_protected:
            return
        mprotect_readonly(self.ptr)
        self.readonly_protected = 1

    cdef writeto(self, out):
        if out is None:
            raise ValueError("Output cannot be None")
        cdef SafeWriter w = SafeWriter(out)
        return w.write(self)

    @classmethod
    def read_from(cls, reader, size_t size):
        if reader is None:
            raise ValueError("reader cannot be None")
        cdef SafeMemory mem = cls(size)
        if size == 0:
            mem.mark_readonly()
            return mem
        cdef SafeReader r = SafeReader(reader)
        cdef size_t n = r.readinto(mem)
        if n < size:
            raise ValueError(f"Expected to read {size} bytes, but got {n} bytes")
        mem.mark_readonly()
        return mem

    @classmethod
    def from_buffer(cls, const unsigned char[:] data):
        if data is None:
            raise ValueError("data cannot be None")
        cdef SafeMemory mem = cls(len(data))
        if len(data) == 0:
            mem.mark_readonly()
            return mem
        mem.set(data)
        return mem


cdef mprotect_readonly(void *ptr):
    if cyd_mprotect_readonly(ptr) != 0:
        raise OSError("Failed to change memory protection to read-only")


cdef uint8_t* malloc_key(size_t size) noexcept nogil:
    return <uint8_t*>cyd_malloc(size)


cdef void free_key(uint8_t* ptr) noexcept nogil:
    cyd_free(ptr)


cdef key_is_zero(const unsigned char[:] key):
    if key is None:
        raise ValueError("key cannot be None")
    cdef size_t lenk = len(key)
    if lenk == 0:
        return True
    return cyd_is_zero(&key[0], lenk) == 1


cdef uint64_t _load64(const unsigned char[:] src) noexcept nogil:
    with cython.boundscheck(False):
        return cyd_load64_be(&src[0])


cdef void _store64(unsigned char[:] dst, uint64_t src) noexcept nogil:
    with cython.boundscheck(False):
        cyd_store64_be(&dst[0], src)


cdef uint32_t _load32(const unsigned char[:] src) noexcept nogil:
    with cython.boundscheck(False):
        return cyd_load32_be(&src[0])


cdef void _store32(unsigned char[:] dst, uint32_t src) noexcept nogil:
    with cython.boundscheck(False):
        cyd_store32_be(&dst[0], src)


cdef uint16_t _load16(const unsigned char[:] src) noexcept nogil:
    with cython.boundscheck(False):
        return cyd_load16_be(&src[0])


cdef void _store16(unsigned char[:] dst, uint16_t src) noexcept nogil:
    with cython.boundscheck(False):
        cyd_store16_be(&dst[0], src)


cpdef load64(const unsigned char[:] src):
    if len(src) < 8:
        raise ValueError(f"src must be 8 bytes long, got {len(src)} bytes")
    return _load64(src)


cpdef store64(unsigned char[:] dst, uint64_t src):
    if len(dst) < 4:
        raise ValueError(f"dst must be 8 bytes long, got {len(dst)} bytes")
    _store64(dst, src)


cpdef load32(const unsigned char[:] src):
    if len(src) < 4:
        raise ValueError(f"src must be 4 bytes long, got {len(src)} bytes")
    return _load32(src)


cpdef store32(unsigned char[:] dst, uint32_t src):
    if len(dst) < 4:
        raise ValueError(f"dst must be 4 bytes long, got {len(dst)} bytes")
    _store32(dst, src)


cpdef load16(const unsigned char[:] src):
    if len(src) < 2:
        raise ValueError(f"src must be 2 bytes long, got {len(src)} bytes")
    return _load16(src)


cpdef store16(unsigned char[:] dst, uint16_t src):
    if len(dst) < 2:
        raise ValueError(f"dst must be 2 bytes long, got {len(dst)} bytes")
    _store16(dst, src)


def have_mman():
    return cyd_have_mman() == 1


def little_endian():
    return cyd_is_little_endian() == 1


def big_endian():
    return cyd_is_big_endian() == 1


cdef class FileOpener:
    def __init__(self, fileobj_or_path, *, mode="rb"):
        if fileobj_or_path is None:
            raise ValueError("fileobj_or_path cannot be None")
        if mode not in ("rb", "wb", "ab"):
            raise ValueError("mode must be 'rb', 'wb', or 'ab'")
        self.fileobj = None
        self.path = None
        self.mode = mode
        if isinstance(fileobj_or_path, (str, os.PathLike)):
            self.path = pathlib.Path(fileobj_or_path)
            return
        if isinstance(fileobj_or_path, (io.IOBase, SafeReader, SafeWriter, tempfile._TemporaryFileWrapper)):
            self.fileobj = fileobj_or_path
            return
        raise TypeError("fileobj must be path-like or a file-like")

    cdef __enter__(self):
        if self.path is not None:
            self.fileobj = open(self.path, self.mode)
        return self.fileobj

    def __exit__(self, exc_type, exc_value, traceback):
        if self.path is not None:
            self.fileobj.close()


cdef class SafeReader:
    def __init__(self, fileobj):
        if fileobj is None:
            raise ValueError("fileobj cannot be None")
        if not hasattr(fileobj, "read"):
            raise TypeError("fileobj must be a file-like object with a 'read' method")
        self.fileobj = fileobj
        self.direct = isinstance(fileobj, (SafeReader, io.BufferedReader, io.BytesIO, io.BufferedRandom))
        self.has_readinto = hasattr(fileobj, "readinto")

    cpdef readinto(self, unsigned char[:] buf):
        if len(buf) == 0:
            return 0
        if self.direct and self.has_readinto:
            return self.fileobj.readinto(buf)
        cdef bytes tmp = self.read(len(buf))
        if len(tmp) == 0:
            return 0
        cdef const unsigned char[:] view = tmp
        buf[0:len(tmp)] = view
        return len(tmp)

    cpdef read(self, size_t length=io.DEFAULT_BUFFER_SIZE):
        if length == 0:
            return b""
        if self.direct == 1:
            return self.fileobj.read(length)

        cdef bytearray result = bytearray(length)
        cdef size_t offset = 0
        cdef bytes tmp
        cdef const unsigned char[:] view

        while offset < length:
            tmp = self.fileobj.read(length - offset)
            if len(tmp) == 0:
                return bytes(result[:offset])
            view = tmp
            result[offset:offset + len(tmp)] = view
            offset += len(tmp)

        return bytes(result[0:offset])


cdef class SafeWriter:
    def __init__(self, fileobj):
        if fileobj is None:
            raise ValueError("fileobj cannot be None")
        if not hasattr(fileobj, "write"):
            raise TypeError("fileobj must be a file-like object with a 'write' method")
        self.fileobj = fileobj
        self.direct = isinstance(fileobj, (SafeWriter, io.BufferedWriter, io.BytesIO, io.BufferedRandom))

    cpdef write(self, const unsigned char[:] buf):
        if buf is None:
            raise ValueError("buf cannot be None")
        if self.direct == 1:
            return self.fileobj.write(buf)
        cdef size_t length = len(buf)
        cdef size_t offset = 0
        cdef size_t n = 0
        while offset < length:
            n = self.fileobj.write(buf[offset:length])
            if n == 0:
                return offset
            offset += n
        return length


cdef class TeeWriter:
    def __init__(self, w1, w2):
        self.w1 = SafeWriter(w1)
        self.w2 = SafeWriter(w2)

    cpdef write(self, const unsigned char[:] buf):
        cdef size_t n1 = self.w1.write(buf)
        cdef size_t n2 = self.w2.write(buf)
        if n1 != n2:
            raise IOError("Writers did not write the same number of bytes")
        return n1
